////////////////////////////////////////////////////////////////////////////////
//                                                                             /
// 2012-2020 (c) Baical                                                        /
//                                                                             /
// This library is free software; you can redistribute it and/or               /
// modify it under the terms of the GNU Lesser General Public                  /
// License as published by the Free Software Foundation; either                /
// version 3.0 of the License, or (at your option) any later version.          /
//                                                                             /
// This library is distributed in the hope that it will be useful,             /
// but WITHOUT ANY WARRANTY; without even the implied warranty of              /
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           /
// Lesser General Public License for more details.                             /
//                                                                             /
// You should have received a copy of the GNU Lesser General Public            /
// License along with this library.                                            /
//                                                                             /
////////////////////////////////////////////////////////////////////////////////
#ifndef ULIST_H
#define ULIST_H

typedef void*(*fnListMalloc)(void *i_pCtx, size_t i_szMem);
typedef void (*fnListFree)(void *i_pCtx, void *i_pMem);
typedef void (*fnListDataFree)(void *i_pCtx, void *i_pMem);

/**
 * \brief list element structure
*/
struct stuEl
{
    void         *pData;  /**< User data */
    struct stuEl *pNext;  /**< Next element */
    struct stuEl *pPrev;  /**< Prev. element */
};


/**
 * \brief list memory manager structure
*/
struct stuListMemMgr
{             
    void        *pCtx;    /**<  Memory manager context*/
    fnListMalloc pMalloc; /**<  Malloc function pointer*/
    fnListFree   pFree;   /**<  Free function pointer*/
};

/**
 * \brief list data manager structure
*/
struct stuListDataMgr
{
    void          *pCtx;  /**<  Data manager context*/
    fnListDataFree pFree; /**<  Free function pointer*/
};


/**
 * \brief list pool structure
*/
struct stuPoolChunk
{
    struct stuEl        *pEls;  /**<  Pool elements array*/
    struct stuPoolChunk *pNext; /**<  Next pool pointer*/
};

/**
 * \brief list structure
*/
struct stuList
{
    struct stuListMemMgr  stMemMgr;      /**<  Memory manager*/
    struct stuListDataMgr stDataMgr;     /**<  Data manager*/

    struct stuEl         *pFirst;        /**<  First element*/
    struct stuEl         *pLast;         /**<  Last element*/
                        
    size_t                szCount;       /**<  Count elements*/
                        
    struct stuEl        **ppIndex;       /**<  Elements index*/
    bool                  bBrokenIndex;  /**<  Is index is broken flag*/
    size_t                szIndexSize;   /**<  Index size*/
    bool                  bInitialized;  /**<  Is list initialized*/
                        
    struct stuPoolChunk  *pPoolFirst;    /**<  First pool chunk*/
    struct stuEl         *pPoolElFirst;  /**<  Fist free list element in pool*/
    size_t                szPoolSize;    /**<  Pool size*/
};

/**
 * \brief create list instance
     * @param i_pMemMgr [in] memory manager context and callbacks
     * @param i_pDataMgr [in] data manager context and callbacks
     * @param i_szPoolSize [in] size of the pool element
     * @return list item
*/
struct stuList* uListCreate(struct stuListMemMgr *i_pMemMgr, struct stuListDataMgr *i_pDataMgr, size_t i_szPoolSize)
{
    if (    (!i_pMemMgr)
         || (!i_pDataMgr)
       )
    {
        return NULL;
    }

    struct stuList *l_pList = (struct stuList *)i_pMemMgr->pMalloc(i_pMemMgr->pCtx, sizeof(struct stuList));

    if (!l_pList)
    {
        return NULL;
    }

    l_pList->bBrokenIndex = true;
    l_pList->bInitialized = true;
    l_pList->szCount      = 0;
    l_pList->szIndexSize  = 0;
    l_pList->pFirst       = NULL;
    l_pList->pLast        = NULL;
    l_pList->stMemMgr     = *i_pMemMgr;
    l_pList->stDataMgr    = *i_pDataMgr;
    l_pList->ppIndex      = NULL;
    l_pList->pPoolFirst   = NULL;
    l_pList->pPoolElFirst = NULL;
    l_pList->szPoolSize   = i_szPoolSize;

    return l_pList;
}


/**
 * \brief release list index table
     * @param i_pList [in] list
*/
static void uListReleaseIndex(struct stuList* i_pList)
{
    if (i_pList)
    {
        if (i_pList->ppIndex)
        {
            i_pList->stMemMgr.pFree(i_pList->stMemMgr.pCtx, i_pList->ppIndex);
            i_pList->ppIndex = NULL;
        }

        i_pList->szIndexSize = 0;
        i_pList->bBrokenIndex = true;
    }
}

/**
 * \brief release pool item
     * @param i_pList [in] list
     * @param io_pPool [in] pool item
*/
static void uListReleasePoolItem(struct stuList* i_pList, struct stuPoolChunk *io_pPool)
{
    if (    (i_pList)
         && (io_pPool)
       )
    {
        if (io_pPool->pEls)
        {
            i_pList->stMemMgr.pFree(i_pList->stMemMgr.pCtx, io_pPool->pEls);
            io_pPool->pEls = NULL;
        }

        i_pList->stMemMgr.pFree(i_pList->stMemMgr.pCtx, io_pPool);
    }
}


/**
 * \brief alloc additional pool item
     * @param i_pList [in] list
*/
static bool uListAllocPoolItem(struct stuList* i_pList)
{
    bool l_bReturn = false;

    if (!i_pList)
    {
        return false;
    }

    struct stuPoolChunk *l_pPool = (struct stuPoolChunk*)i_pList->stMemMgr.pMalloc(i_pList->stMemMgr.pCtx,sizeof(struct stuPoolChunk));
        
    if (l_pPool)
    {
        memset(l_pPool, 0, sizeof(struct stuPoolChunk));

        l_pPool->pEls = (struct stuEl*)i_pList->stMemMgr.pMalloc(i_pList->stMemMgr.pCtx, sizeof(struct stuEl) * i_pList->szPoolSize);
            
        if (l_pPool->pEls)
        {
            memset(l_pPool->pEls, 0, sizeof(struct stuPoolChunk) * i_pList->szPoolSize);

            struct stuEl *l_pEl = l_pPool->pEls;

            for (size_t l_dwIDX = 1; l_dwIDX < i_pList->szPoolSize; l_dwIDX++)
            {
                l_pEl->pNext = (l_pEl + 1);
                l_pEl ++;
            }

            l_bReturn             = true;
            l_pPool->pNext        = i_pList->pPoolFirst;
            i_pList->pPoolFirst   = l_pPool;
            l_pEl->pNext        = i_pList->pPoolElFirst;
            i_pList->pPoolElFirst = l_pPool->pEls;
        }
    }

    if (false == l_bReturn)
    {
        uListReleasePoolItem(i_pList, l_pPool);
    }

    return l_bReturn;
}


/**
 * \brief free list element
     * @param i_pList [in] list
     * @param i_pEl [in] list element
     * @param i_bFreeEl [in] free list element flag
     * @param i_bFreeData [in] free user data flag
*/
static inline bool uListFreeElement(struct stuList* i_pList, struct stuEl *i_pEl, bool i_bFreeEl, bool i_bFreeData)
{
    if (    (!i_pList)
         || (!i_pEl)
       )
    {
        return false;
    }

    bool l_bResult = true;

    if ((NULL != i_pEl->pPrev) && (NULL != i_pEl->pNext)) 
    {
        i_pEl->pPrev->pNext = i_pEl->pNext;
        i_pEl->pNext->pPrev = i_pEl->pPrev;
    }
    else
    { 
        l_bResult = false;

        if (i_pEl == i_pList->pFirst)
        {
            i_pList->pFirst = i_pEl->pNext;

            if (i_pList->pFirst != 0)
            {
                i_pList->pFirst->pPrev = 0;
            }

            l_bResult = true;  
        }

        if (i_pEl == i_pList->pLast)
        {
            i_pList->pLast = i_pEl->pPrev;
            if (i_pList->pLast != 0) 
            {
                i_pList->pLast->pNext = 0; 
            }

            l_bResult = true;  
        }
    }

    if (l_bResult) 
    {
        if (i_bFreeData)
        {
            i_pList->stDataMgr.pFree(i_pList->stDataMgr.pCtx, i_pEl->pData);
            i_pEl->pData = 0;
        }

        if (i_bFreeEl)
        {
            memset(i_pEl, 0, sizeof(struct stuEl));
            i_pEl->pNext = i_pList->pPoolElFirst;
            i_pList->pPoolElFirst = i_pEl;
        }
    }

    i_pList->bBrokenIndex = true;

    i_pList->szCount--;

    return l_bResult;
}


/**
 * \brief rebuild list index
     * @param i_pList [in] list
     * @return bool
*/
static bool uListBuildIndex(struct stuList* i_pList)
{
    bool l_bResult = true;

    if (!i_pList)
    {
        return false;
    }

    if (    (i_pList->ppIndex)
         && (i_pList->szCount > i_pList->szIndexSize) 
       )
    {
        uListReleaseIndex(i_pList);
    }

    if (NULL == i_pList->ppIndex)
    {
        i_pList->szIndexSize = i_pList->szCount + 128;
        i_pList->ppIndex = (struct stuEl**)i_pList->stMemMgr.pMalloc(i_pList->stMemMgr.pCtx, 
                                                                       sizeof(struct stuEl*) * i_pList->szIndexSize);
    }

    if (i_pList->ppIndex)
    {
        struct stuEl *l_pEl = i_pList->pFirst;

        memset(i_pList->ppIndex, 0, i_pList->szIndexSize * sizeof(struct stuEl*));
            
        struct stuEl **l_pIndexLocal = i_pList->ppIndex;

        while (l_pEl)
        {
            *l_pIndexLocal = l_pEl;
            l_pIndexLocal ++;
            l_pEl = l_pEl->pNext;
        }

        i_pList->bBrokenIndex = false;
    }
    else
    {
        l_bResult = false;
        i_pList->szIndexSize = 0;
    }

    return l_bResult;
}



/**
 * \brief get first element
     * @param i_pList [in] list
     * @return first element
*/
inline struct stuEl *uListGetFirst(struct stuList* i_pList)
{
    if (!i_pList)
    {
        return NULL;
    }

    return i_pList->pFirst;
}


/**
 * \brief get last element
     * @param i_pList [in] list
     * @return last element
*/
inline struct stuEl *uListGetLast(struct stuList* i_pList)
{
    if (!i_pList)
    {
        return NULL;
    }

    return i_pList->pLast;
}


/**
 * \brief get list size in elements
     * @param i_pList [in] list
     * @return size
*/
inline size_t uListSize(struct stuList* i_pList)
{
    if (!i_pList)
    {
        return 0;
    }

    return i_pList->szCount;
}


/**
 * \brief remove list element
     * @param i_pList [in] list
     * @param i_pEl [in] list element
     * @param i_bFree_Data [in] free user data flag
     * @return bool
*/
bool uListDel(struct stuList* i_pList, struct stuEl *i_pEl, bool i_bFree_Data)
{
    if (    (!i_pList)
         || (!i_pEl)
       )
    {
        return false;
    }

    return uListFreeElement(i_pList, i_pEl, true, i_bFree_Data);
}

/**
 * \brief extract element from the list
     * @param i_pList [in] list
     * @param i_pEl [in] list element
     * @return bool
*/
bool uListExtract(struct stuList* i_pList, struct stuEl *i_pEl)
{
    if (    (!i_pList)
         || (!i_pEl)
       )
    {
        return false;
    }

    bool l_bResult = uListFreeElement(i_pList, i_pEl, false, false);
    i_pEl->pNext = NULL;
    i_pEl->pPrev = NULL;

    return l_bResult;
}


/**
 * \brief remove first element from the list
     * @param i_pList [in] list
     * @return first element user data
*/
void *uListPullFirst(struct stuList* i_pList)
{
    if (    (!i_pList)
         || (!i_pList->pFirst)
       )
    {
        return NULL;
    }

    void *l_pReturn = i_pList->pFirst->pData;
    uListFreeElement(i_pList, i_pList->pFirst, true, false);
    return l_pReturn;
}

/**
 * \brief remove last element from the list
     * @param i_pList [in] list
     * @return first element user data
*/
void *uListPullLast(struct stuList* i_pList)
{
    if (    (!i_pList)
         || (!i_pList->pLast)
       )
    {
        return NULL;
    }

    void *l_pReturn = i_pList->pLast->pData;
    uListFreeElement(i_pList, i_pList->pLast, true, false);
    return l_pReturn;
}


/**
 * \brief allocate new list element
     * @param i_pList [in] list
     * @return new element
*/
static struct stuEl *uListAllocateEl(struct stuList* i_pList)
{
    if (!i_pList)
    {
        return NULL;
    }

    struct stuEl *l_pReturn = NULL;
    if (NULL == i_pList->pPoolElFirst)
    {
        uListAllocPoolItem(i_pList);
    }

    l_pReturn = i_pList->pPoolElFirst;

    if (i_pList->pPoolElFirst)
    {
        i_pList->pPoolElFirst = i_pList->pPoolElFirst->pNext;
    }

    return l_pReturn;
}


/**
 * \brief insert new list element after i_pListEl
     * @param i_pList [in] list
     * @param i_pListEl [in] existing list element, NULL means first
     * @param i_pNewEl [in] new element
     * @return bool
*/
bool uListPutAfter(struct stuList* i_pList, struct stuEl *i_pListEl, struct stuEl *i_pNewEl)
{
    if (    (!i_pList) 
         || (!i_pNewEl)
       )
    {
        return false;
    }

    bool l_bResult = true;

    i_pList->szCount++;

    if (i_pListEl)
    {
        i_pNewEl->pNext = i_pListEl->pNext;
        i_pNewEl->pPrev = i_pListEl;
        if (i_pListEl->pNext)
        {
            i_pListEl->pNext->pPrev = i_pNewEl;
        }
        i_pListEl->pNext = i_pNewEl;
    }


    if ((NULL == i_pListEl) && (i_pList->pFirst))
    {
        i_pList->pFirst->pPrev = i_pNewEl;
        i_pNewEl->pNext = i_pList->pFirst;
        i_pList->pFirst = i_pNewEl;
    }

    if ((i_pList->pLast) && (i_pList->pLast == i_pListEl))
    {
        i_pList->pLast = i_pNewEl;
    }

    if (!i_pList->pFirst)
    {
        i_pList->pFirst = i_pNewEl;
        i_pList->pLast  = i_pNewEl;
    }

    i_pList->bBrokenIndex = true;
     
    return l_bResult;
}


/**
 * \brief add new element after existing element
     * @param i_pList [in] list
     * @param i_pEl [in] list element after which new data will be inserted
     * @param i_pData [in] user data
     * @return new element
*/
static inline struct stuEl *uListAddAfter(struct stuList* i_pList, struct stuEl *i_pEl, void *i_pData)
{
    if (!i_pList)
    {
        return false;
    }

    struct stuEl *l_pNewEl = uListAllocateEl(i_pList);

    if (l_pNewEl)
    {
        memset(l_pNewEl, 0, sizeof(struct stuEl));
        l_pNewEl->pData = i_pData;
        l_pNewEl->pNext = NULL;
        l_pNewEl->pPrev = NULL;

        uListPutAfter(i_pList, i_pEl, l_pNewEl);
    }

    return l_pNewEl;
}//CListBase::Add_After


/**
 * \brief Add list item at first place
     * @param i_pList [in] list
     * @param i_pData [in] user data
     * @return list element
*/
struct stuEl *uListPushFirst(struct stuList* i_pList, void *i_pData)
{
    return uListAddAfter(i_pList, NULL, i_pData);
}


/**
 * \brief Add list item at last place
     * @param i_pList [in] list
     * @param i_pData [in] user data
     * @return list element
*/
struct stuEl *uListPushLast(struct stuList* i_pList, void *i_pData)
{
    if (!i_pList)
    {
        return NULL;
    }

    return uListAddAfter(i_pList, i_pList->pLast, i_pData);
}


/**
 * \brief Get list prev item
     * @param i_pList [in] list
     * @param i_pEl [in] list element
     * @return prev list element
*/
struct stuEl *uListGetPrev(struct stuList* i_pList, struct stuEl *i_pEl)
{
    struct stuEl *l_pResult = NULL;

    if (!i_pList)
    {
        return NULL;
    }

    if (i_pEl)
    {
        l_pResult = i_pEl->pPrev;
    }
    else
    {
        l_pResult = i_pList->pLast;
    }

    return l_pResult;
}


/**
 * \brief Get list next item
     * @param i_pList [in] list
     * @param i_pEl [in] list element
     * @return next list element
*/
struct stuEl *uListGetNext(struct stuList* i_pList, struct stuEl *i_pEl)
{
    struct stuEl *l_pResult = NULL;

    if (!i_pList)
    {
        return NULL;
    }

    if (i_pEl)
    {
        l_pResult = i_pEl->pNext;
    }
    else
    {
        l_pResult = i_pList->pFirst;
    }

    return l_pResult;
}


/**
 * \brief Get list item user data
     * @param i_pList [in] list
     * @param i_pEl [in] list element
     * @return user data
*/
void *uListGetData(struct stuList* i_pList, struct stuEl *i_pEl)
{
    if (    (!i_pList)
         || (!i_pEl)
       )
    {
        return NULL;
    }

    return i_pEl->pData;
}


/**
 * \brief Replace list item user data by new one
     * @param i_pList [in] list
     * @param i_pEl [in] list element
     * @param i_bFreeOldData [in] release/free old data flag
     * @param i_pData [in] new user data
     * @return user data
*/
bool uListSetData(struct stuList* i_pList, struct stuEl *i_pEl, bool i_bFreeOldData, void *i_pData)
{
    if (    (!i_pList)
         || (!i_pEl)
       )
    {
        return false;
    }

    if (    (i_bFreeOldData)
         && (i_pEl->pData)
       )
    {
        i_pList->stDataMgr.pFree(i_pList->stDataMgr.pCtx, i_pEl->pData);
    }

    i_pEl->pData = i_pData;

    return true;
}


/**
 * \brief Get list element by index
     * @param i_pList [in] list
     * @param i_szIndex [in] list element index starting from 0
     * @return list element
*/
struct stuEl *uListGetByIndex(struct stuList* i_pList, size_t i_szIndex)
{
    struct stuEl *l_pReturn = NULL;
    uint32_t      l_dwI     = 0;

    if (!i_pList)
    {
        return NULL;
    }

    if (i_pList->bBrokenIndex)
    {
        uListBuildIndex(i_pList);
    }

    if  (i_szIndex < i_pList->szCount) 
    {
        if (    (false == i_pList->bBrokenIndex)
             && (i_pList->ppIndex) 
           )
        {
            l_pReturn = i_pList->ppIndex[i_szIndex];
        }
        else
        {
            l_pReturn = i_pList->pFirst;
            for (l_dwI = 0; l_dwI < i_szIndex; l_dwI++)
            {
                if (l_pReturn)
                {
                    l_pReturn = l_pReturn->pNext;
                }
                else
                {
                    break;
                }
            }
        }
    }

    return l_pReturn;
}//CListBase::Get_ByIndex


/**
 * \brief release list instance
     * @param i_pList [in] release list
     * @param i_bClearData [in] clear list & user data flag
*/
void uListFree(struct stuList* i_pList, bool i_bClearData)
{
    if (!i_pList)
    {
        return;
    }

    uListReleaseIndex(i_pList);
    
    if (i_bClearData)
    {
        while (NULL != i_pList->pFirst)
        {
            if (true != uListFreeElement(i_pList, i_pList->pFirst, true, i_bClearData))
            {
                break;
            }
        }
    }
    
    struct stuPoolChunk *l_pPool_Tmp = NULL;
    while (i_pList->pPoolFirst)
    {
        l_pPool_Tmp = i_pList->pPoolFirst;
        i_pList->pPoolFirst = i_pList->pPoolFirst->pNext;
    
        uListReleasePoolItem(i_pList, l_pPool_Tmp);
    }

    i_pList->stMemMgr.pFree(i_pList->stMemMgr.pCtx, i_pList);
}

#endif //ULIST_H
