////////////////////////////////////////////////////////////////////////////////
//                                                                             /
// 2012-2020 (c) Baical                                                        /
//                                                                             /
// This library is free software; you can redistribute it and/or               /
// modify it under the terms of the GNU Lesser General Public                  /
// License as published by the Free Software Foundation; either                /
// version 3.0 of the License, or (at your option) any later version.          /
//                                                                             /
// This library is distributed in the hope that it will be useful,             /
// but WITHOUT ANY WARRANTY; without even the implied warranty of              /
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           /
// Lesser General Public License for more details.                             /
//                                                                             /
// You should have received a copy of the GNU Lesser General Public            /
// License along with this library.                                            /
//                                                                             /
////////////////////////////////////////////////////////////////////////////////
//******************************************************************************
// This header file contains red black tree pooled implementation              *
// - http://www.rsdn.ru/article/alg/binstree.xml                               *
// - http://algolist.manual.ru/ds/rbtree.php                                   *
// - http://rain.ifmo.ru/cat/view.php/vis/trees/red-black-2002                 *
//******************************************************************************
#ifndef URBTREE_H
#define URBTREE_H


#define RBT_NODE_RED                                                         (1)
#define RBT_NODE_BLACK                                                       (0)

typedef void*(*fnTreeMalloc)(void *i_pCtx, size_t i_szMem);
typedef void (*fnTreeFree)(void *i_pCtx, void *i_pMem);
typedef void (*fnTreeDataFree)(void *i_pCtx, void *i_pMem);

/**
 * \brief Compare user key with tree data
     * @param i_pUserKey [in] user key
     * @param i_pTreeData [in] tree data, inside has to be key copy
     * @return if (i_pUserKey < i_pTreeData->pKey) true else false
*/
typedef bool(* fnTreeKeyLess)(void *i_pUserKey, void *i_pTreeData);

/**
 * \brief Compare user key with tree data
     * @param i_pUserKey [in] user key
     * @param i_pTreeData [in] tree data, inside has to be key copy
     * @return if (i_pUserKey == i_pTreeData->pKey) true else false
*/
typedef bool(* fnTreeKeysEqual)(void *i_pUserKey, void *i_pTreeData);


/**
 * \brief Tree node description
*/
struct stuTreeNode
{
    struct stuTreeNode *pLeft;     /**<  Left branch*/
    struct stuTreeNode *pRight;    /**<  Right branch*/
    struct stuTreeNode *pParent;   /**<  Parent*/
    void               *tData;     /**<  User data*/
    char                bColor;    /**<  Color: black/red*/
};

/**
 * \brief Tree pool chunk description
*/
struct stuTreePoolChunk
{
    struct stuTreeNode      *pNodes;   /**<  Allocated nodes array*/
    size_t                   szCount;  /**<  Count*/
    struct stuTreePoolChunk *pNext;    /**<  Pointer to next chunk*/ 
};

/**
 * \brief list memory manager structure
*/
struct stuTreeMemMgr
{
    void        *pCtx;    /**<  Memory manager context*/
    fnTreeMalloc pMalloc; /**<  Malloc function pointer*/
    fnTreeFree   pFree;   /**<  Free function pointer*/
};

/**
 * \brief list data manager structure
*/
struct stuTreeDataMgr
{
    void           *pCtx;       /**<  Data manager context*/
    fnTreeDataFree  pFree;      /**<  Free function pointer*/
    fnTreeKeyLess   pKeyLess;   /**<  Keys comparison function*/
    fnTreeKeysEqual pKeysEqual; /**<  Keys comparison function*/
};


/**
 * \brief tree structure
*/
struct stuTree
{
    struct stuTreeMemMgr     stMemMgr;      /**<  Memory manager*/
    struct stuTreeDataMgr    stDataMgr;     /**<  Data manager*/
    struct stuTreePoolChunk *pPoolFirst;    /**<  Pool first chunk pointer*/ 
    struct stuTreeNode      *pPoolFirstNode;/**<  Pool first free node pointer*/
    size_t                   szPoolSize;    /**<  Pool default size */
    struct stuTreeNode       stSentinel;    /**<  Empty node*/
    struct stuTreeNode      *pRoot;         /**<  Root node*/
    struct stuTreeNode      *pNIL;          /**<  NULL node*/
};


/**
 * \brief create list instance
     * @param i_pMemMgr [in] memory manager context and callbacks
     * @param i_pDataMgr [in] data manager context and callbacks
     * @param i_szPoolSize [in] size of the pool element
     * @return list item
*/
struct stuTree* uTreeCreate(struct stuTreeMemMgr  *i_pMemMgr, 
                            struct stuTreeDataMgr *i_pDataMgr, 
                            size_t                 i_szPoolSize
                           )
{
    if (    (!i_pMemMgr)
         || (!i_pDataMgr)
       )
    {
        return NULL;
    }

    struct stuTree *l_pTree = (struct stuTree *)i_pMemMgr->pMalloc(i_pMemMgr->pCtx, sizeof(struct stuTree));

    if (!l_pTree)
    {
        return NULL;
    }

    l_pTree->pNIL               = &l_pTree->stSentinel;
    l_pTree->pRoot              = l_pTree->pNIL;    

    l_pTree->stSentinel.bColor  = RBT_NODE_BLACK;
    l_pTree->stSentinel.pParent = NULL;
    l_pTree->stSentinel.tData   = NULL;
    l_pTree->stSentinel.pLeft   = l_pTree->pNIL;
    l_pTree->stSentinel.pRight  = l_pTree->pNIL;

    l_pTree->szPoolSize         = i_szPoolSize;
    l_pTree->pPoolFirst         = NULL;
    l_pTree->pPoolFirstNode     = NULL;

    l_pTree->stMemMgr           = *i_pMemMgr;
    l_pTree->stDataMgr          = *i_pDataMgr;

    return l_pTree;
}


/**
 * \brief Release pool chunk, private function
     * @param i_pTree [in] tree pointer
     * @param io_pPool [in] pool chunk
*/
static void uTreeFreePoolChunk(struct stuTree *i_pTree, struct stuTreePoolChunk *io_pPool)
{
    if (!i_pTree)
    {
        return;
    }

    if (io_pPool)
    {
        if (io_pPool->pNodes)
        {
            i_pTree->stMemMgr.pFree(i_pTree->stMemMgr.pCtx, io_pPool->pNodes);
            io_pPool->pNodes = NULL;
        }
        i_pTree->stMemMgr.pFree(i_pTree->stMemMgr.pCtx, io_pPool);
    }
}


/**
 * \brief Allocate pool chunk, private function
     * @param i_pTree [in] tree pointer
     * @return bool
*/
static bool uTreeAllocPoolChunk(struct stuTree *i_pTree)
{
    bool l_bReturn = false;

    if (!i_pTree)
    {
        return false;
    }

    struct stuTreePoolChunk *l_pPool = (struct stuTreePoolChunk*)i_pTree->stMemMgr.pMalloc(i_pTree->stMemMgr.pCtx, sizeof(struct stuTreePoolChunk));
        
    if (l_pPool)
    {
        memset(l_pPool, 0, sizeof(struct stuTreePoolChunk));
        l_pPool->szCount = i_pTree->szPoolSize;
        l_pPool->pNodes  = (struct stuTreeNode*)i_pTree->stMemMgr.pMalloc(i_pTree->stMemMgr.pCtx, sizeof(struct stuTreeNode) * l_pPool->szCount);
            
        if (l_pPool->pNodes)
        {
            struct stuTreeNode *l_pNode = NULL;

            memset(l_pPool->pNodes, 
                   0, 
                   sizeof(struct stuTreePoolChunk) * l_pPool->szCount
                  );

            l_pNode = l_pPool->pNodes;

            for (size_t l_szI = 1; l_szI < l_pPool->szCount; l_szI++)
            {
                l_pNode->pRight = (l_pNode + 1);
                l_pNode ++;
            }

            l_bReturn                   = true;
            l_pPool->pNext            = i_pTree->pPoolFirst;
            i_pTree->pPoolFirst      = l_pPool;
            l_pNode->pRight             = i_pTree->pPoolFirstNode;
            i_pTree->pPoolFirstNode = l_pPool->pNodes;
        }
    }

    if (!l_bReturn)
    {
        uTreeFreePoolChunk(i_pTree, l_pPool);
    }

    return l_bReturn;
}

/**
 * \brief Allocate tree node, private function
     * @param i_pTree [in] tree pointer
     * @return bool
*/
static struct stuTreeNode *uTreeAllocNode(struct stuTree *i_pTree)
{
    struct stuTreeNode *l_pReturn = NULL;
    if (NULL == i_pTree->pPoolFirstNode)
    {
        uTreeAllocPoolChunk(i_pTree);
    }

    l_pReturn = i_pTree->pPoolFirstNode;

    if (i_pTree->pPoolFirstNode)
    {
        i_pTree->pPoolFirstNode = i_pTree->pPoolFirstNode->pRight;
    }

    return l_pReturn;
}

/**
 * \brief Rotate sub nodes in left, private function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to be rotated
*/
static void uTreeRotateLeft(struct stuTree *i_pTree, struct stuTreeNode *i_pNode) 
{
    if (!i_pTree)
    {
        return;
    }

    struct stuTreeNode *l_pNodeR = i_pNode->pRight;

    i_pNode->pRight = l_pNodeR->pLeft;
    if (i_pTree->pNIL != l_pNodeR->pLeft) 
    {
        l_pNodeR->pLeft->pParent = i_pNode;
    }

    if (i_pTree->pNIL != l_pNodeR) 
    {
        l_pNodeR->pParent = i_pNode->pParent;
    }

    if (i_pNode->pParent) 
    {
        if (i_pNode == i_pNode->pParent->pLeft)
        {
            i_pNode->pParent->pLeft = l_pNodeR;
        }
        else
        {
            i_pNode->pParent->pRight = l_pNodeR;
        }
    } 
    else 
    {
        i_pTree->pRoot = l_pNodeR;
    }

    l_pNodeR->pLeft = i_pNode;
    if (i_pTree->pNIL != i_pNode) 
    {
        i_pNode->pParent = l_pNodeR;
    }
}


/**
 * \brief Rotate sub nodes in right, private function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to be rotated
*/
static void uTreeRotateRight(struct stuTree *i_pTree, struct stuTreeNode *i_pNode) 
{
    if (!i_pTree)
    {
        return;
    }

    struct stuTreeNode *l_pNodeL = i_pNode->pLeft;

    i_pNode->pLeft = l_pNodeL->pRight;
    if (i_pTree->pNIL != l_pNodeL->pRight)
    {
        l_pNodeL->pRight->pParent = i_pNode;
    }

    if (i_pTree->pNIL != l_pNodeL) 
    {
        l_pNodeL->pParent = i_pNode->pParent;
    }

    if (i_pNode->pParent) 
    {
        if (i_pNode == i_pNode->pParent->pRight)
        {
            i_pNode->pParent->pRight = l_pNodeL;
        }
        else
        {
            i_pNode->pParent->pLeft = l_pNodeL;
        }
    }
    else 
    {
        i_pTree->pRoot = l_pNodeL;
    }

    l_pNodeL->pRight = i_pNode;

    if (i_pTree->pNIL != i_pNode)
    {
        i_pNode->pParent = l_pNodeL;
    }
}


/**
 * \brief Repair tree after delete, private function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to be rotated
*/
static void uTreeRepairDel(struct stuTree *i_pTree, struct stuTreeNode *i_pNode) 
{
    if (!i_pTree)
    {
        return;
    }

    if (!i_pTree)
    {
        return;
    }

    while (    (i_pNode != i_pTree->pRoot)
            && (RBT_NODE_BLACK == i_pNode->bColor)
          )
    {
        if (i_pNode == i_pNode->pParent->pLeft) 
        {
            struct stuTreeNode *l_pBrother = i_pNode->pParent->pRight;
            if (RBT_NODE_RED == l_pBrother->bColor)
            {
                l_pBrother->bColor       = RBT_NODE_BLACK;
                i_pNode->pParent->bColor = RBT_NODE_RED;
                uTreeRotateLeft(i_pTree, i_pNode->pParent);
                l_pBrother = i_pNode->pParent->pRight;
            }

            if (    (RBT_NODE_BLACK == l_pBrother->pLeft->bColor) 
                 && (RBT_NODE_BLACK == l_pBrother->pRight->bColor) 
               )
            {
                l_pBrother->bColor = RBT_NODE_RED;
                i_pNode            = i_pNode->pParent;
            } 
            else 
            {
                if (RBT_NODE_BLACK == l_pBrother->pRight->bColor)
                {
                    l_pBrother->pLeft->bColor = RBT_NODE_BLACK;
                    l_pBrother->bColor        = RBT_NODE_RED;
                    uTreeRotateRight (i_pTree, l_pBrother);
                    l_pBrother = i_pNode->pParent->pRight;
                }
                l_pBrother->bColor = i_pNode->pParent->bColor;
                i_pNode->pParent->bColor   = RBT_NODE_BLACK;
                l_pBrother->pRight->bColor = RBT_NODE_BLACK;
                uTreeRotateLeft (i_pTree, i_pNode->pParent);
                i_pNode = i_pTree->pRoot;
            }
        } 
        else
        {
            struct stuTreeNode *l_pBrother = i_pNode->pParent->pLeft;
            if (RBT_NODE_RED == l_pBrother->bColor) 
            {
                l_pBrother->bColor       = RBT_NODE_BLACK;
                i_pNode->pParent->bColor = RBT_NODE_RED;
                uTreeRotateRight (i_pTree, i_pNode->pParent);
                l_pBrother = i_pNode->pParent->pLeft;
            }
            if (    (RBT_NODE_BLACK == l_pBrother->pRight->bColor)
                 && (RBT_NODE_BLACK == l_pBrother->pLeft->bColor) 
               )
            {
                l_pBrother->bColor = RBT_NODE_RED;
                i_pNode            = i_pNode->pParent;
            } 
            else
            {
                if (RBT_NODE_BLACK == l_pBrother->pLeft->bColor)
                {
                    l_pBrother->pRight->bColor = RBT_NODE_BLACK;
                    l_pBrother->bColor         = RBT_NODE_RED;
                    uTreeRotateLeft (i_pTree, l_pBrother);
                    l_pBrother = i_pNode->pParent->pLeft;
                }
                l_pBrother->bColor        = i_pNode->pParent->bColor;
                i_pNode->pParent->bColor  = RBT_NODE_BLACK;
                l_pBrother->pLeft->bColor = RBT_NODE_BLACK;
                uTreeRotateRight (i_pTree, i_pNode->pParent);
                i_pNode = i_pTree->pRoot;
            }
        }
    }

    i_pNode->bColor = RBT_NODE_BLACK;
}


/**
 * \brief Repair tree after add, private function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to be rotated
*/
static void uTreeRepairPush(struct stuTree *i_pTree, struct stuTreeNode *i_pNode) 
{
    if (!i_pTree)
    {
        return;
    }

    while (    (i_pNode != i_pTree->pRoot)
            && (RBT_NODE_RED == i_pNode->pParent->bColor) 
          )
    {
        // we have a violation
        if (i_pNode->pParent == i_pNode->pParent->pParent->pLeft) 
        {
            struct stuTreeNode *l_pUncle = i_pNode->pParent->pParent->pRight;
            if (RBT_NODE_RED == l_pUncle->bColor) 
            {
                i_pNode->pParent->bColor          = RBT_NODE_BLACK;
                l_pUncle->bColor                  = RBT_NODE_BLACK;
                i_pNode->pParent->pParent->bColor = RBT_NODE_RED;
                i_pNode                           = i_pNode->pParent->pParent;
            } 
            else //(RBT_NODE_BLACK == l_pUncle->bColor) 
            {
                if (i_pNode == i_pNode->pParent->pRight) 
                {
                    // make x a left child 
                    i_pNode = i_pNode->pParent;
                    uTreeRotateLeft(i_pTree, i_pNode);
                }

                // recolor and rotate 
                i_pNode->pParent->bColor          = RBT_NODE_BLACK;
                i_pNode->pParent->pParent->bColor = RBT_NODE_RED;
                uTreeRotateRight(i_pTree, i_pNode->pParent->pParent);
            }
        }
        else //(i_pNode->pParent == i_pNode->pParent->pParent->pRight) 
        {
            struct stuTreeNode *l_pUncle = i_pNode->pParent->pParent->pLeft;
            if (RBT_NODE_RED == l_pUncle->bColor) 
            {
                i_pNode->pParent->bColor          = RBT_NODE_BLACK;
                l_pUncle->bColor                  = RBT_NODE_BLACK;
                i_pNode->pParent->pParent->bColor = RBT_NODE_RED;
                i_pNode                           = i_pNode->pParent->pParent;
            } 
            else //(RBT_NODE_BLACK == l_pUncle->bColor) 
            {
                if (i_pNode == i_pNode->pParent->pLeft) 
                {
                    i_pNode = i_pNode->pParent;
                    uTreeRotateRight(i_pTree, i_pNode);
                }

                i_pNode->pParent->bColor          = RBT_NODE_BLACK;
                i_pNode->pParent->pParent->bColor = RBT_NODE_RED;
                uTreeRotateLeft(i_pTree, i_pNode->pParent->pParent);
            }
        }
    }

    i_pTree->pRoot->bColor = RBT_NODE_BLACK;
}//Fix_Push


/**
 * \brief Free node, private function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to be rotated
*/
static void uTreeFreeNode(struct stuTree *i_pTree, struct stuTreeNode *i_pNode)
{
    if (!i_pTree)
    {
        return;
    }

    memset(i_pNode, 0, sizeof(struct stuTreeNode));
    i_pNode->pRight             = i_pTree->pPoolFirstNode;
    i_pTree->pPoolFirstNode = i_pNode;
}


/**
 * \brief Delete node, private function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to be rotated
*/
static void uTreeDeleteNode(struct stuTree *i_pTree, struct stuTreeNode *i_pNode)
{
    if (!i_pTree)
    {
        return;
    }

    if (i_pTree->pNIL != i_pNode->pLeft)
    {
        uTreeDeleteNode(i_pTree, i_pNode->pLeft);
    }

    if (i_pTree->pNIL != i_pNode->pRight)
    {
        uTreeDeleteNode(i_pTree, i_pNode->pRight);
    }

    if (i_pTree->stDataMgr.pFree)
    {
        i_pTree->stDataMgr.pFree(i_pTree->stDataMgr.pCtx, i_pNode->tData);
    }

    i_pNode->tData = NULL;

    if (i_pNode->pParent) 
    {
        if (i_pNode == i_pNode->pParent->pRight)
        {
            i_pNode->pParent->pRight = i_pTree->pNIL;
        }
        else
        {
            i_pNode->pParent->pLeft = i_pTree->pNIL;
        }
    }

    uTreeFreeNode(i_pTree, i_pNode);
}

/**
 * \brief Add node, public function
     * @param i_pTree [in] tree pointer
     * @param i_pData [in] user data to be inserted
     * @param i_pKey [in] user key (part of the data)
     * @return bool
*/
bool uTreePush(struct stuTree *i_pTree, void *i_pData, void *i_pKey) 
{
    if (!i_pTree)
    {
        return false;
    }

    struct stuTreeNode *l_pCurrent = i_pTree->pRoot; 
    struct stuTreeNode *l_pParent  = NULL; 
    struct stuTreeNode *l_pNode    = NULL;


    //find where node belongs
    while (i_pTree->pNIL != l_pCurrent)
    {
        if (i_pTree->stDataMgr.pKeysEqual(i_pKey, l_pCurrent->tData))
        {
            return true;
        }
        else
        {
            l_pParent = l_pCurrent;
            l_pCurrent = i_pTree->stDataMgr.pKeyLess(i_pKey, l_pCurrent->tData) ? l_pCurrent->pLeft : l_pCurrent->pRight;
        }
    }

    l_pNode = uTreeAllocNode(i_pTree);

    if (NULL == l_pNode) 
    {
        return false;
    }

    l_pNode->tData   = i_pData;
    l_pNode->pParent = l_pParent;
    l_pNode->pLeft   = i_pTree->pNIL;
    l_pNode->pRight  = i_pTree->pNIL;
    l_pNode->bColor  = RBT_NODE_RED;

    // insert node in tree 
    if (l_pParent)
    {
        if (i_pTree->stDataMgr.pKeyLess(i_pKey, l_pParent->tData))
        {
            l_pParent->pLeft = l_pNode;
        }
        else
        {
            l_pParent->pRight = l_pNode;
        }
    }
    else 
    {
        i_pTree->pRoot = l_pNode;
    }

    uTreeRepairPush(i_pTree, l_pNode);
    return true;
}


/**
 * \brief Find node value by key, public function
     * @param i_pTree [in] tree pointer
     * @param i_pKey [in] user key
     * @return node data 
*/
void *uTreeFind(struct stuTree *i_pTree, void *i_pKey) 
{
    if (!i_pTree)
    {
        return NULL;
    }

    struct stuTreeNode *l_pCurrent = i_pTree->pRoot;
    while (i_pTree->pNIL != l_pCurrent)
    {
        if (i_pTree->stDataMgr.pKeysEqual(i_pKey, l_pCurrent->tData))
        {
            return (l_pCurrent->tData);
        }
        else
        {
            l_pCurrent =   i_pTree->stDataMgr.pKeyLess(i_pKey, l_pCurrent->tData) 
                         ? l_pCurrent->pLeft
                         : l_pCurrent->pRight;
        }
    }

    return NULL;
}


/**
 * \brief Find node by key, public function
     * @param i_pTree [in] tree pointer
     * @param i_pKey [in] user key
     * @return node data 
*/
struct stuTreeNode *uTreeFindNode(struct stuTree *i_pTree, void *i_pKey) 
{
    if (!i_pTree)
    {
        return NULL;
    }

    struct stuTreeNode *l_pCurrent = i_pTree->pRoot;
    while (i_pTree->pNIL != l_pCurrent)
    {
        if (i_pTree->stDataMgr.pKeysEqual(i_pKey, l_pCurrent->tData))
        {
            return l_pCurrent;
        }
        else
        {
            l_pCurrent =   i_pTree->stDataMgr.pKeyLess(i_pKey, l_pCurrent->tData) 
                         ? l_pCurrent->pLeft
                         : l_pCurrent->pRight;
        }
    }

    return NULL;
}


/**
 * \brief Get node depth, private function
     * @param i_pTree [in] tree pointer
     * @param o_pMin [out] min depth
     * @param o_pMax [out] max depth
     * @param i_dwCurrent [in] current depth
*/
void uTreeGetNodeDepth(struct stuTree     *i_pTree,
                       struct stuTreeNode *i_pNode, 
                       uint64_t           *o_pMin, 
                       uint64_t           *o_pMax,
                       uint64_t            i_dwCurrent
                      )
{
    if (!i_pTree)
    {
        return;
    }

    i_dwCurrent ++;

    if (    (i_pTree->pNIL == i_pNode->pLeft)
         && (i_pTree->pNIL == i_pNode->pRight)
       )
    {
        if (i_dwCurrent < (*o_pMin))
        {
            *o_pMin = i_dwCurrent;
        }

        if (i_dwCurrent > (*o_pMax))
        {
            *o_pMax = i_dwCurrent;
        }
    }
    else
    {
        if (i_pTree->pNIL != i_pNode->pLeft)
        {
            uTreeGetNodeDepth(i_pTree, i_pNode->pLeft, o_pMin, o_pMax, i_dwCurrent);
        }

        if (i_pTree->pNIL != i_pNode->pRight)
        {
            uTreeGetNodeDepth(i_pTree, i_pNode->pRight, o_pMin, o_pMax, i_dwCurrent);
        }
    }
}


/**
 * \brief Get depth, public function
     * @param i_pTree [in] tree pointer
     * @param o_pMin [out] min depth
     * @param o_pMax [out] max depth
*/
void uTreeGetDepth(struct stuTree *i_pTree, uint64_t *o_pMin, uint64_t *o_pMax)
{
    if (    (!o_pMin)
         || (!o_pMax)
         || (!i_pTree)
       )
    {
        return;
    }

    *o_pMin = ((uint64_t)~((uint64_t)0)); //max possible value
    *o_pMax = 0;

    if (i_pTree->pNIL != i_pTree->pRoot)
    {
        uTreeGetNodeDepth(i_pTree, i_pTree->pRoot, o_pMin, o_pMax, 0);
    }
}
 

/**
 * \brief Delete node, public function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to delete
*/
void uTreeDel(struct stuTree *i_pTree, struct stuTreeNode *i_pNode) 
{
    struct stuTreeNode *l_pX = NULL;
    struct stuTreeNode *l_pY = NULL;

    if (    (NULL == i_pNode)
         || (!i_pTree)
         || (i_pTree->pNIL  == i_pNode)
        )
    {
        return;
    }

    if (i_pTree->stDataMgr.pFree)
    {
        i_pTree->stDataMgr.pFree(i_pTree->stDataMgr.pCtx, i_pNode->tData);
        i_pNode->tData = NULL;
    }

    if (    (i_pTree->pNIL == i_pNode->pLeft)
         || (i_pTree->pNIL == i_pNode->pRight) 
       )
    {
        //y has a m_pNIL node as a child
        l_pY = i_pNode;
    } 
    else 
    {
        //find tree successor with a m_pNIL node as a child
        l_pY = i_pNode->pRight;
        while (i_pTree->pNIL != l_pY->pLeft)
        {
            l_pY = l_pY->pLeft;
        }
    }

    //x is y's only child
    if (i_pTree->pNIL != l_pY->pLeft)
    {
        l_pX = l_pY->pLeft;
    }
    else
    {
        l_pX = l_pY->pRight;
    }

    // remove y from the parent chain 
    l_pX->pParent = l_pY->pParent;
    if (l_pY->pParent)
    {
        if (l_pY == l_pY->pParent->pLeft)
        {
            l_pY->pParent->pLeft = l_pX;
        }
        else
        {
            l_pY->pParent->pRight = l_pX;
        }
    }
    else
    {
        i_pTree->pRoot = l_pX;
    }

    if (l_pY != i_pNode) 
    {
        i_pNode->tData = l_pY->tData;
    }

    if (l_pY->bColor == RBT_NODE_BLACK)
    {
        uTreeRepairDel(i_pTree, l_pX);
    }

    uTreeFreeNode(i_pTree, l_pY);
}//Del


/**
 * \brief Clear tree, public function
     * @param i_pTree [in] tree pointer
     * @param i_pNode [in] node to delete
*/
void uTreeClear(struct stuTree *i_pTree)
{
    if (!i_pTree)
    {
        return;
    }

    if (i_pTree->pNIL == i_pTree->pRoot)
    {
        return;
    }

    uTreeDeleteNode(i_pTree, i_pTree->pRoot);
    i_pTree->pRoot = i_pTree->pNIL;
}//Clear


void uTreeFree(struct stuTree *i_pTree)
{
    if (!i_pTree)
    {
        return;
    }

    //delete all tree
    uTreeClear(i_pTree);

    struct stuTreePoolChunk *l_pPool_Tmp = NULL;
    while (i_pTree->pPoolFirst)
    {
        l_pPool_Tmp         = i_pTree->pPoolFirst;
        i_pTree->pPoolFirst = i_pTree->pPoolFirst->pNext;

        uTreeFreePoolChunk(i_pTree, l_pPool_Tmp);
    }

    i_pTree->stMemMgr.pFree(i_pTree->stMemMgr.pCtx, i_pTree);
}


#endif //URBTREE_H
